from enum import Enum
import copy

class CellularAutomata:
    """
    A class representing a basic cellular automata using a boolean function as a rule.
    """

    def __init__(self, func, initialState):
        self.func = func
        self.state = []
        for cell in initialState:
            if cell == 1:
                self.state.append(func.field.ONE)
            elif cell == 0:
                self.state.append(func.field.ZERO)
            else:
                self.state.append(cell)

        self.time = []

    def update(self):
        """
        Updates this Cellular Automata to the next state, and appends the old state to the time
        attribute.
        """
        self.time.append(copy.deepcopy(self.state))
        out = []
        for i in range(0, len(self.state)):
            start = i - (self.func.n//2)
            end = (i + (self.func.n//2)) % len(self.state) + 1
            out.append(self.func(*_circle_slice(self.state, start, end)))

        self.state = out
        
         
    def pretty_print(self, colorMap=None):
        """
        Prints the state over time in a format that is easily visible in a Bash command line.
        """
        colorMap = colorMap if colorMap is not None else {self.func.field.ZERO: " ", self.func.field.ONE: "\033[7m \033[0m"} 
        for row in self.time:
            for cell in row:
                print(colorMap[cell], sep="", end="")

            print()

    def get_column(self, col):
        """
        Gets a column over time of the CA.

        Args:
            col (int): An integer representing the column to get

        Returns:
            list: The column over time, with index 0 being the initial state.
        """
        return [row[col] for row in self.time]

class TargetTransitions(Enum):
        """
            An Enum representing which states under which a target CA is allowed to change state.
        """
        TARGET_TO_TARGET = 0
        TARGET_TO_NON = 1
        NON_TO_TARGET = 2
        NON_TO_NON = 3
        
        def standard_revs():
            """
            Returns the standard revision values, allowing moving from Target state to Target State,
            Nontarget state to Nontarget state, and Nontarget State to Target State.

            Returns:
                TargetTransitions: The above states.
            """
            return TargetTransitions.TARGET_TO_TARGET, TargetTransitions.NON_TO_TARGET, TargetTransitions.NON_TO_NON

class TargetedCellularAutomata(CellularAutomata):
    def __init__(self, func, initialState, targetState, *args):
        super().__init__(func, initialState)
        self.targetState = targetState
        self.reversionValues = frozenset(args)

    def update(self):
        super().update()
        oldState = self.time[-1]
        for i in range(len(oldState)):
            change = _get_change_state(self.targetState[i], oldState[i], self.state[i])
            if change not in self.reversionValues:
                self.state[i] = oldState[i]

class FiniteStateMachine:
    """
    A class representing a finite state automata for a boolean function.
    """    

    def __init__(self, func):
        self.n = func.n
        self.func = func
        self.__generateGraph()

    def __generateGraph(self):
        self.adjMatr = []
        vertices = [_dec_to_bin(i, self.n - 1) for i in range(0, 2**(self.n - 1))]
        for i in range(0, (2**(self.n-1))):
            thisRow = []
            for j in vertices:
                a = _dec_to_bin(i, self.n - 1) 
                if a[1:] == j[:len(j) - 1]:
                    thisRow.append(self.func([a[0]] + j[:]))
                else:
                    thisRow.append(None)
            self.adjMatr.append(thisRow)

def _circle_slice(lst, start, end):
    """
    Slices a list in a loop. Treats the list as a ring and slices from beginning to end,
    looping back to the beginning of the list if necessary.
    """
    if 0 <= start < end < len(lst):
        return lst[start:end]
    elif start < 0:
        return lst[start+len(lst):] + lst[:end]
    elif end >= len(lst):
        return lst[start:] + lst[0:end-len(lst)]
    elif start > end:
        return lst[start:] + lst[:end]
    elif start == end:
        return lst[start:] + lst[:end]

    print("SLICE FAILURE: ", lst, "FROM:", start, "END:", end)
    return []


def _get_change_state(target, curr, nex):
    is_curr_tar = (target == curr)
    is_nex_tar = (target == nex)

    if is_curr_tar and not is_nex_tar:
        return TargetTransitions.TARGET_TO_NON
    elif is_nex_tar and not is_curr_tar:
        return TargetTransitions.NON_TO_TARGET
    elif is_nex_tar and is_curr_tar:
        return TargetTransitions.TARGET_TO_TARGET
    else:
        return TargetTransitions.NON_TO_NON



from setuptools import setup

setup(name="booleantools",
      version="0.3.2",
      description="A library for boolean functions, functions on GF(4), and Cellular Automata",
      url=None,
      author="Wesley Rogers, Andrew Penland",
      author_email="wsrogers3@catamount.wcu.edu",
      license="MIT",
      python_requires=">=3",
      packages = ["booleantools"],
      #py_modules=["booleantools"],
      zip_safe=False)



import booleantools as bt
import unittest

r30 = bt.BooleanFunction([[0],[1],[2],[1,2]], 3)
r90 = bt.BooleanFunction([[0],[2]], 3)
x = bt.getX(5)

class BTUnitTest(unittest.TestCase):
    
    def test_hamming_wt(self):
        self.assertEqual(r30.hamming_weight(), 4)

    def test_ham_dist(self):
        self.assertEqual(r30.hamming_distance(r90), 2)
        self.assertEqual(r30.hamming_distance([r90, r30]), [2, 0])
        
    def test_nonlinearity(self):
        self.assertEqual(r30.nonlinearity(), 2)
        self.assertEqual(r90.nonlinearity(), 0)

    def test_eval(self):
        r30p1 = bt.BooleanFunction([[0],[1],[2],[1,2],[]], 3)
        self.assertEqual(r30(1,0,0), bt.GF2.get(1))
        self.assertEqual(r30(0,0,0), bt.GF2.get(0))
        self.assertEqual(r30p1(0,0,0), bt.GF2.get(1))

        self.assertEqual(r90(1,0,1), bt.GF2.get(0))
        self.assertEqual(r90(1,1,0), bt.GF2.get(1))

    def test_perm(self):
        self.assertEqual(r30.apply_permutation([0,1,2]), r30)
        self.assertNotEqual(r30.apply_permutation([1,0,2]), r30)
        
    def test_balanced(self):
        self.assertTrue(r30.is_balanced())

    def test_correlation_immune(self):
        self.assertFalse(r30.is_correlation_immune())
        self.assertTrue(r90.is_correlation_immune())
    
    def test_linear_structs(self):
        self.assertEqual(r30.linear_structures(), set([0]))
        self.assertEqual(r90.linear_structures(), set([0,2]))
    
    def test_str(self):
        self.assertEqual(repr(r30), "BooleanFunction([[0], [1], [2], [1, 2]], 3)")
        self.assertEqual(str(r30), "x_{0} \\oplus x_{1} \\oplus x_{2} \\oplus x_{1}x_{2}")
        self.assertEqual(r30.tex_str(True), "$x_{0} \\oplus x_{1} \\oplus x_{2} \\oplus x_{1}x_{2}$")
        
    def test_add(self):
        rule = bt.BooleanFunction([[1],[1,2]], 3)
        self.assertEqual(r30 + r90, rule)

    def test_hash(self):
        self.assertEqual(hash(r30), 30)
        self.assertEqual(hash(r90), 90)

    def test_k_resilient(self):
        self.assertFalse(r30.is_k_resilient())
        self.assertTrue(r90.is_k_resilient())
        self.assertFalse(r90.is_k_resilient(k=2))
    
    def test_hausdorff_dist(self):
        r30_class = r30.get_orbit()
        r90_class = r90.get_orbit()
        self.assertEqual(bt.hausdorff_distance(r30_class, r90_class), 2)
    
    def test_mult(self):
        prod = bt.BooleanFunction([[0],[2],[0,1],[0,1,2]], 3)
        self.assertEqual(prod, r30*r90)
    
    def test_generate_function(self):
        new_r30 = bt.generate_function(30, 3)
        new_r90 = bt.generate_function(90, 3)
        self.assertEqual(r30, new_r30)
        self.assertEqual(r90, new_r90)
    
    def test_siegenthaler(self):
        expected_comb = bt.BooleanFunction([[0], [1],[2],[1,2],[1,3],[1,2,3]], 4)
        got = bt.siegenthaler_combination(r90, r30, x[3])
        self.assertEqual(expected_comb, got)

if __name__ == "__main__":
    unittest.main()
